/*	Application		IPC receiver-side using socket (UDP)		
	File			IPC_socket_Receiver.cpp
	Purpose			Demonstration of local IPC, using a blocking socket
	Author			Richard John Anthony
	Date			December 2013
	
	Function		1. Creates a socket, and binds to local address (uses INADDR_ANY)
					2. Waits for a message to arrive on its port (8007)
					3. Displays the message and then exits once a key is pressed
*/

#include "stdafx.h"
#include <iostream>
#include <atlsocket.h>
#include "IPC_socket_Receiver.h"

using namespace std;

void main()
{
	cout << "Local Inter-Process Communication (receiver side)" << endl;
    cout << "This program waits for a message to arrive on port 8007 (decimal)." << endl;
    cout << "The message is displayed and then the process exits when a key is pressed." << endl;
    cout << "--------------------------------------------------------------------------" << endl;
    cout << "Waiting for message ...   ";

	// Initialise the Windows socket mechanism
	bool bSuccess;
	bSuccess = InitialiseWindowsSocketsMechanism();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Windows socket mechanism could not be initialised" << endl;
		CloseSocketAndExit();
	}
	
	// Create a socket - this is the logical 'communication endpoint'
	bSuccess = CreateSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}

	// Set the Local Address details (IP address and  port number)
	bSuccess = SetUpLocalAddressStruct();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Local address could not be retrieved" << endl;
		CloseSocketAndExit();
	}

	// Bind (The operating system makes a mapping between this process and the requested Port)
	bSuccess = BindToLocalAddress();
	if(false == bSuccess)
	{
		cout << "Exiting the program because binding to the requested port was not allowed (port may be in use)" << endl;
		CloseSocketAndExit();
	}

	// Wait for / Receive the message
	ReadMessageFromReceiveBufferOrWait();
	cout << "Message received:   " << m_szRecvBuf << endl;

	// Close the socket
	CloseSocketAndExit();
}
	
bool InitialiseWindowsSocketsMechanism(void)
{
	WSADATA wsaData;
	int iResult = WSAStartup( MAKEWORD(2,2), &wsaData );  // Initialize Winsock
	if ( iResult != NO_ERROR )
	{
		cout << "Error at WSAStartup()" << endl;
		return false;
	}
	return true;
}

bool CreateSocket(void)
{
	m_IPC_ReceiveSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
		// Create a socket and retain a reference to it
		// AF_INET indicates that Internet addressing will be used
		// SOCK_DGRAM indicates that discrete messages will be used
		// PF_INET indicates that Internet protocols will be used
	if(INVALID_SOCKET == m_IPC_ReceiveSOCKET)
	{
		cout << "CreateSocket() failed" << endl;	// Report error if socket could not be created
		return false;
	}
	return true;
}

bool SetUpLocalAddressStruct(void)
{
	// Initialise a SOCKADDR_IN structure with the local address (for receiving)
	m_LocalSockAddr.sin_addr.S_un.S_addr = htonl(INADDR_ANY);
		// INADDR_ANY indicates that a message will be received if it was sent to ANY
		// of the local computer's IP addresses (it may have one or more)
	m_LocalSockAddr.sin_family = AF_INET;
	m_LocalSockAddr.sin_port = htons(8007); // Set the port number, which will be used to 
											// map an incomming message to this process
	return true;
}

bool BindToLocalAddress(void)
{
	// This is where the process associates its socket and thus itself with a port, using the bind() primitive.
	// The operating system makes an association between the process and the port number it binds to.
	// From this point on, the operating system will deliver messages addressed by the selected port number, to this process
	int iError = bind(m_IPC_ReceiveSOCKET,(const SOCKADDR FAR*)&m_LocalSockAddr, sizeof(m_LocalSockAddr));
	if(SOCKET_ERROR == iError)
	{
		cout << "bind() Failed!";	// Report error if bind failed (port may already be in use)
		return false;
	}
	return true;
}

void ReadMessageFromReceiveBufferOrWait(void)
{
	// The process uses the receivefrom() to begin waiting for a message.
	// If a message is already in the receive buffer it will be delivered immediately to the process.
	// If a message has not yet arrived, the process will wait 
	// (the operating system will move it to 'blocked' state).
	int iBytesRecd = recvfrom(m_IPC_ReceiveSOCKET, (char FAR*)m_szRecvBuf, RECEIVE_BUFFER_SIZE, 0, NULL, NULL);
	if(SOCKET_ERROR == iBytesRecd)
	{
		cout << "Receive failed" << endl;	// Report error if receive operation failed
		CloseSocketAndExit();
	}
	else
	{	
		m_szRecvBuf[iBytesRecd] = 0; // Ensure null termination of the message string in the buffer
	}
}

void CloseSocketAndExit()
{
	char CUserResponse;
 	if(0 != m_IPC_ReceiveSOCKET)
	{
		closesocket(m_IPC_ReceiveSOCKET);	// If the socket was open, Close the socket
	}
	WSACleanup();	
	cout << "Socket closed" << endl << endl;
	cout << "(press a key to continue)" << endl;
	cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
	exit(0);
}