/*	Application		IPC receiver-side using socket (UDP)		
	File			IPC_socket_Receiver_with_Reply.cpp
	Purpose			Demonstration of local IPC, using a blocking socket
	Author			Richard John Anthony
	Date			December 2013
	
	Function		1. Creates a socket, and binds to local address (uses INADDR_ANY)
					2. Waits for a message to arrive on its port (8007)
					3. Displays the message 
					4. Reverses the message contents
					5. Sends the modified message back as a reply to the sender (to port 8008)
					6. Exits once a key is pressed
*/

#include "stdafx.h"
#include <iostream>
#include <atlsocket.h>
#include "IPC_socket_Receiver_with_Reply.h"

using namespace std;

void main()
{
	cout << "Local Inter-Process Communication (receiver side)" << endl;
    cout << "This program waits for a message to arrive on port 8007 (decimal)." << endl;
    cout << "The message is displayed, and the message contents are reversed." << endl;
	cout << "The modified message is sent back to the sender as a reply, on port 8008." << endl;
	cout << "The process exits when a key is pressed." << endl;
    cout << "--------------------------------------------------------------------------" << endl;
    cout << "Waiting for message ...   " << endl;

	// Initialise the Windows socket mechanism
	bool bSuccess;
	bSuccess = InitialiseWindowsSocketsMechanism();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Windows socket mechanism could not be initialised" << endl;
		CloseSocketAndExit();
	}
	
	// Create a socket - this is the logical 'communication endpoint'
	bSuccess = CreateReceiveSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}

	// Create the socket that will be used for sending the reply message
	bSuccess = CreateReplySocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}

	// Set the Local Address details (IP address and  port number)
	bSuccess = SetUpLocalAddressStruct();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Local address could not be retrieved" << endl;
		CloseSocketAndExit();
	}

	// Bind (The operating system makes a mapping between this process and the requested Port)
	bSuccess = BindToLocalAddress();
	if(false == bSuccess)
	{
		cout << "Exiting the program because binding to the requested port was not allowed (port may be in use)" << endl;
		CloseSocketAndExit();
	}

	// Wait for / Receive the message
	ReadMessageFromReceiveBufferOrWait();
	cout << "   Message received:       " << m_szRecvBuf << endl;
	cout << "   from sender at address: " << m_szSender_IP_Address << endl;
	
	// Reverse the received message contents in the buffer
	ReverseBufferContents();
	cout << "   Reversed reply message: " << m_szReplyBuf << endl;

	// Send the reply message to the original message sender
	SendMessage();
	cout << "   Reply message sent" << endl << endl;
	
	// The following short delay is only necessary due to the very simple design of the demonstration programs. In particular, the sender uses a bloc
	Sleep(500);	// Wait a short while to ensure the sender process (which will receive the reply) has time to set up its receive system call

	// Close the socket
	CloseSocketAndExit();
}
	
bool InitialiseWindowsSocketsMechanism(void)
{
	WSADATA wsaData;
	int iResult = WSAStartup( MAKEWORD(2,2), &wsaData );  // Initialize Winsock
	if ( iResult != NO_ERROR )
	{
		cout << "Error at WSAStartup()" << endl;
		return false;
	}
	return true;
}

bool CreateReceiveSocket(void)
{
	m_IPC_ReceiveSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_IPC_ReceiveSOCKET)
	{
		cout << "CreateSocket() failed" << endl;
		return false;
	}
	return true;
}

bool CreateReplySocket(void)
{
	m_IPC_ReplySOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_IPC_ReplySOCKET)
	{
		cout << "CreateSocket() failed" << endl;
		return false;
	}
	return true;
}

bool SetUpLocalAddressStruct(void)
{
	m_LocalSockAddr.sin_addr.S_un.S_addr = htonl(INADDR_ANY);
	m_LocalSockAddr.sin_family = AF_INET;
	m_LocalSockAddr.sin_port = htons(8007); // Decimal
	return true;
}

bool BindToLocalAddress(void)
{
	// This is where the process associates itself with a port to receive on
	int iError = bind(m_IPC_ReceiveSOCKET,(const SOCKADDR FAR*)&m_LocalSockAddr, 
					sizeof(m_LocalSockAddr));
	if(SOCKET_ERROR == iError)
	{
		cout << "bind() Failed!";
		return false;
	}
	return true;
}

void ReadMessageFromReceiveBufferOrWait(void)
{
	int iSenderSockAddr_Length = sizeof(m_SenderSockAddr);
	
	// The recvfrom call is modified by providing the address and length of a SOCKADDR_IN structure 
	// to hold the address of the message sender (so a reply can be sent back).
	int iBytesRecd = recvfrom(m_IPC_ReceiveSOCKET, (char FAR*)m_szRecvBuf, RECEIVE_BUFFER_SIZE, 
					0, (SOCKADDR FAR*)&m_SenderSockAddr, &iSenderSockAddr_Length);
	if(SOCKET_ERROR == iBytesRecd)
	{
		cout << "Receive failed" << endl;
		CloseSocketAndExit();
	}
	else
	{	
		m_szRecvBuf[iBytesRecd] = 0; // Ensure null termination
		sprintf(m_szSender_IP_Address, "%d.%d.%d.%d",
			m_SenderSockAddr.sin_addr.S_un.S_un_b.s_b1,
			m_SenderSockAddr.sin_addr.S_un.S_un_b.s_b2,
			m_SenderSockAddr.sin_addr.S_un.S_un_b.s_b3,	
			m_SenderSockAddr.sin_addr.S_un.S_un_b.s_b4);
	}
}

void ReverseBufferContents()
{
	int iLength = strlen(m_szRecvBuf);
	for(int iIndex = 0; iIndex < iLength; iIndex++)
	{
		m_szReplyBuf[(iLength - 1) - iIndex] = m_szRecvBuf[iIndex];
	}
	int i = 4;
}

void SendMessage(void)
{
	int iLength = strlen(m_szReplyBuf);
	
	// Set the reply port to be the same as the one the sender will use to receive on
	m_SenderSockAddr.sin_port = htons(8008); // Decimal; 

	// Send a UDP datagram containing the repy message
	int iBytesSent = sendto(m_IPC_ReplySOCKET, (char FAR *)&m_szReplyBuf, iLength, 0,
		(const struct sockaddr FAR *)&m_SenderSockAddr, sizeof(m_SenderSockAddr));
	if(INVALID_SOCKET == iBytesSent)
	{
		cout << "sendto() Failed!" << endl;
		CloseSocketAndExit();
	}
}

void CloseSocketAndExit()
{
	char CUserResponse;
 	if(0 != m_IPC_ReceiveSOCKET)
	{
		closesocket(m_IPC_ReceiveSOCKET);	// If the socket was open, Close the socket
	}
 	if(0 != m_IPC_ReplySOCKET)
	{
		closesocket(m_IPC_ReplySOCKET);	// If the socket was open, Close the socket
	}
	WSACleanup();	
	cout << "Socket closed" << endl << endl;
	cout << "(press a key to continue)" << endl;
	cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
	exit(0);
}