/*	Application		IPC sender-side using socket (UDP)		
	File			IPC_socket_Sender_With_reply.cpp
	Purpose			Demonstration of local IPC, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			December 2013
	
	Function		1. Creates a socket
					2. Waits for user to type a message
					3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
					4. Waits for a reply message to be returned (on port 8008)
					5. Displays the reply message
					6. Exits once the reply message has been displayed
*/

#include "stdafx.h"
#include <iostream>
#include <string>
#include <atlsocket.h>

using namespace std;

#include "IPC_socket_Sender_With_reply.h"

void main()
{
	int iMessageLength = 0;
	m_UserInput = "";
	cout << "Local Inter-Process Communication (sender side)" << endl;
	cout << "This program waits for the user to type a message," << endl;
	cout << "sends it to port 8007 on the local computer, and" << endl;
	cout << "waits for a reply message to be sent back on port 8008." << endl;
	cout << "------------------------------------------------------" << endl;
	cout << "Type a  message to send ...   ";
	getline(cin, m_UserInput); // Get a string which may include spaces

	// Initialise the Windows socket mechanism
	bool bSuccess;
	bSuccess = InitialiseWindowsSocketsMechanism();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Windows socket mechanism could not be initialised" << endl;
		CloseSocketAndExit();
	}

	// Create Send socket
	bSuccess = CreateSendSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}

	// Create receive-reply socket
	bSuccess = CreateReceiveReplySocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketAndExit();
	}
	
	// Set the Send Address details (Use local loopback IP address (127.0.0.1), and port 8007)
	SetUpSendAddressStructFor_LocalLoopback();
	
	// Set up the local address structure used with bind() so the reply message can be received
	SetUpLocalAddressStruct();

	// Bind to local address and port 8008 so reply message can be received
	BindToLocalAddress();

	// Send the message that the user typed, to the sepcified port on the local computer
	bSuccess = SendMessage();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the message could not be sent" << endl;
		CloseSocketAndExit();
	}
	cout << "Message sent" << endl;

	// Message has ben sent - now wait for the reply
	ReadMessageFromReceiveBufferOrWait();
	cout << "   Reply message received:    " << m_szReceiveBuf << endl << endl;
	
	CloseSocketAndExit();
}
	
bool InitialiseWindowsSocketsMechanism(void)
{
	
	WSADATA wsaData;
	int iResult = WSAStartup( MAKEWORD(2,2), &wsaData );  // Initialize Winsock
	if ( iResult != NO_ERROR )
	{
		cout << "Error at WSAStartup()" << endl;
		return false;
	}
	return true;
}

bool CreateSendSocket(void)
{
	m_SendSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_SendSOCKET)
	{
		cout << "socket() failed" << endl;
		return false;
	}
	return true;
}

bool CreateReceiveReplySocket(void)
{
	m_ReceiveReplySOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_ReceiveReplySOCKET)
	{
		cout << "socket() failed" << endl;
		return false;
	}
	return true;
}

void SetUpSendAddressStructFor_LocalLoopback(void)
{
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b1 = (unsigned  char) 127;
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b2 = (unsigned  char) 0;
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b3 = (unsigned  char) 0;
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b4 = (unsigned  char) 1;
	m_SendSockAddr.sin_family = AF_INET;
	m_SendSockAddr.sin_port = htons(8007);
}

void SetUpLocalAddressStruct(void) // This is needed for bind() so the reply message can be received
{
	m_LocalSockAddr.sin_addr.S_un.S_addr = htonl(INADDR_ANY);
	m_LocalSockAddr.sin_family = AF_INET;
	m_LocalSockAddr.sin_port = htons(8008); // Decimal
}

bool BindToLocalAddress(void)
{
	// This is where the process associates itself with a port to receive on
	int iError = bind(m_ReceiveReplySOCKET,(const SOCKADDR FAR*)&m_LocalSockAddr, 
					sizeof(m_LocalSockAddr));
	if(SOCKET_ERROR == iError)
	{
		cout << "bind() Failed!";
		return false;
	}
	return true;
}

bool SendMessage(void)
{
	m_iSendLen = m_UserInput.length();
	
	// Send a UDP datagram containing the user's typed message
	int iBytesSent = sendto(m_SendSOCKET, (char FAR *)m_UserInput.c_str(), m_iSendLen, 0,
			(const struct sockaddr FAR *)&m_SendSockAddr, sizeof(m_SendSockAddr));
	if(INVALID_SOCKET == iBytesSent)
	{
		cout << "sendto() Failed!" << endl;
		return false;
	}
	return true;
}

void ReadMessageFromReceiveBufferOrWait(void)
{
	int iBytesRecd = recvfrom(m_ReceiveReplySOCKET, (char FAR*)m_szReceiveBuf, RECEIVE_BUFFER_SIZE, 
					0, NULL, NULL);
	if(SOCKET_ERROR == iBytesRecd)
	{
		cout << "Receive failed" << endl;
		CloseSocketAndExit();
	}
	else
	{	
		m_szReceiveBuf[iBytesRecd] = 0; // Ensure null termination
	}
}

void CloseSocketAndExit()
{
	char CUserResponse;
 	if(0 != m_SendSOCKET)
	{
		closesocket(m_SendSOCKET);	// If the socket was open, Close the socket
	}
	WSACleanup();	
	cout << "Socket closed" << endl << endl;
	cout << "(press a key to continue)" << endl;
	cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
	exit(0);
}