/*	Application		Socket-based IPC threads example	
	File			Multithreaded_IPC_two_worker_threads.cpp
	Purpose			Demonstration of local IPC, and multithreading using a blocking socket
	Author			Richard John Anthony
	Date			February 2014
	
	Usage			Requires two command-line arguments {send port, recv port}

	Function		1. Creates a pair of sockets (one for receiving, one for sending)
					2. Binds to the user-provided receive-port number
					3. Starts a worker thread to handle receive activity
					4. Starts a worker thread to handle send activity
					4. Main thread is reserved for application business logic (null in this simple demonstration)
					5. Exits once the string "Quit" (or "QUIT", "quit") is entered
*/

#include "stdafx.h"
#include <iostream>
#include <atlsocket.h>
#include "Multithreaded_IPC_two_worker_threads.h"
#include <thread>

using namespace std;

int main(int argc, char* argv[])
{
	char CUserResponse;
	if(3 != argc)
	{
		// User did not provide the required parameters
		cout << endl << "Usage: " << argv[0] << " SendPort_number ReceivePort_number" << endl;
		cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
		return -1;
	}
	
	m_iSendPort = atoi(argv[1]);
	if(1023 >= m_iSendPort)
	{
		// Send Port value is in reserved port range
		cout << endl << "Send port number must be 1024 or greater (up to 65535)" << endl;
		cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
		return -1;
	}
	
	m_iReceivePort = atoi(argv[2]);
	if(1023 >= m_iReceivePort)
	{
		// Receive port value is in reserved port range
		cout << endl << "Receive port number must be 1024 or greater (up to 65535)" << endl;
		cin >> CUserResponse;	// Wait for user to type a key (to allow them to read the message before the window is closed)
		return -1;
	}

	cout << "Multithreaded Local Inter-Process Communication" << endl;
    cout << "This process sendss messages on port " << m_iSendPort << " and receives messages on port " << m_iReceivePort << endl;
	cout << "Sending is performed by the main thread" << endl;
	cout << "Receiving is performed by a worker thread" << endl;
    cout << "--------------------------------------------------------------------------" << endl;
    
	cout << "Initialising resources" << endl;
	
	// Initialise the Windows socket mechanism
	bool bSuccess;
	bSuccess = InitialiseWindowsSocketsMechanism();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the Windows socket mechanism could not be initialised" << endl;
		CloseSocketsAndExit();
	}
	
	// Create the socket that will be used for sending messages
	bSuccess = CreateSendSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketsAndExit();
	}

	// Create a socket that will be used for receiving messages
	bSuccess = CreateReceiveSocket();
	if(false == bSuccess)
	{
		cout << "Exiting the program because a socket could not be created" << endl;
		CloseSocketsAndExit();
	}

	// Set the Local Address details (IP address and port number)
	bSuccess = SetUpLocalAddressStruct();
	if(false == bSuccess)
	{
		cout << "Exiting the program because the local address could not be retrieved" << endl;
		CloseSocketsAndExit();
	}

	// Bind (The operating system makes a mapping between this process and the requested Port)
	bSuccess = BindToLocalAddress();
	if(false == bSuccess)
	{
		cout << "Exiting the program because binding to the requested port was not allowed (port may be in use)" << endl;
		CloseSocketsAndExit();
	}

	SetUpSendAddressStructFor_LocalLoopback();

	m_bApplicationRunning = true;

	// Create and start worker thread (performs Receive activity in loop)
    thread ReceiveThread(ReceiveMessages);

	// Use detach() to cause the ReceiveThread to run asynchronously with main()
	ReceiveThread.detach();

	// Create and start worker thread (performs Send activity in loop)
    thread SendThread(SendMessages_Thread);

	// Use detach() to cause the SendThread to run asynchronously with main()
	SendThread.detach();

	// The Main thread is reserved for application business logic.
	// In this simple demonstration there is no business logic as such, so the main loop is implememented
	// as a simple loop which repeatedly displays the text " -MainThread- " at 5-second intervals
	while(1)
	{
		if(true == m_bApplicationRunning)
		{
			cout << endl << " -MainThread- " << endl;
			Sleep(5000);	// 5 second delay
		}
	}
	return 0;
}
	
bool InitialiseWindowsSocketsMechanism(void)
{
	WSADATA wsaData;
	int iResult = WSAStartup( MAKEWORD(2,2), &wsaData );  // Initialize Winsock
	if ( iResult != NO_ERROR )
	{
		cout << "Error at WSAStartup()" << endl;
		return false;
	}
	return true;
}

bool CreateSendSocket(void)
{
	m_IPC_SendSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_IPC_SendSOCKET)
	{
		cout << "CreateSocket() failed" << endl;
		return false;
	}
	return true;
}

bool CreateReceiveSocket(void)
{
	m_IPC_ReceiveSOCKET = socket(AF_INET, SOCK_DGRAM, PF_UNSPEC);
	if(INVALID_SOCKET == m_IPC_ReceiveSOCKET)
	{
		cout << "CreateSocket() failed" << endl;
		return false;
	}
	return true;
}

bool SetUpLocalAddressStruct(void)
{
	m_LocalSockAddr.sin_addr.S_un.S_addr = htonl(INADDR_ANY);
	m_LocalSockAddr.sin_family = AF_INET;
	m_LocalSockAddr.sin_port = htons(m_iReceivePort);					// Set the port number (to receive on)
	return true;
}

bool BindToLocalAddress(void)
{	// This is where the process associates itself with a port to receive on

	// Caution, when using namespace std, bind() associates with std::bind() by default - which is the wrong function!
	// Solution is to use the scope resolution operator without a scope name to signify to use the version of bind() 
	// that is in the global name space (which does not have a name). I.e. use ::bind()
	int iError = ::bind(m_IPC_ReceiveSOCKET,(const SOCKADDR FAR*)&m_LocalSockAddr, sizeof(m_LocalSockAddr));
	if(SOCKET_ERROR == iError)
	{
		cout << "bind() Failed!";
		return false;
	}
	return true;
}

void SetUpSendAddressStructFor_LocalLoopback(void)
{   // Initialise a SOCKADDR_IN structure with the recipients socket address details
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b1 = (unsigned  char) 127;	// Set the four bytes of internet address
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b2 = (unsigned  char) 0;		// This is the address of the recipient
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b3 = (unsigned  char) 0;		// The loopback address is used because
	m_SendSockAddr.sin_addr.S_un.S_un_b.s_b4 = (unsigned  char) 1;		// both processes will be on the same computer
	m_SendSockAddr.sin_family = AF_INET;
	m_SendSockAddr.sin_port = htons(m_iSendPort);						// Set the port number (to send to)
}

// ******************************* Beginning of Receive Thread code *******************************
void ReceiveMessages(void) // The Receive Thread body
{
	cout << endl << "ReceiveThread: Waiting for message ...   " << endl;

	while(true == m_bApplicationRunning)
	{
		ReadMessageFromReceiveBufferOrWait();
	}
}

void ReadMessageFromReceiveBufferOrWait(void)
{
	// The recvfrom call is modified by providing the address and length of a SOCKADDR_IN structure 
	// to hold the address of the message sender (so a reply can be sent back).
	int iBytesRecd = recvfrom(m_IPC_ReceiveSOCKET, (char FAR*)m_szRecvBuf, RECEIVE_BUFFER_SIZE, 
					0, NULL, NULL);
	if(true == m_bApplicationRunning)
	{
		if(SOCKET_ERROR == iBytesRecd)
		{
			cout << endl << "ReceiveThread: Receive failed" << endl;
		}
		else
		{	
			m_szRecvBuf[iBytesRecd] = 0; // Ensure null termination
			cout << endl << "ReceiveThread: Message received:       " << m_szRecvBuf << endl;
		}
	}
}
// ********************************* End of Receive Thread code *********************************

// ******************************* Beginning of Send Thread code ********************************
void SendMessages_Thread()
{
	int iNumberOfBytesSent;

	while(1)
	{
		for(int iIndex = 0; iIndex < SEND_BUFFER_SIZE; iIndex++)
		{   // Clear the input buffer
			m_szSendBuf[iIndex] = 0;
		}

		cout << endl << "SendThread: Type a message to send (\"Quit\" to quit): ";
		cin >> m_szSendBuf;
		if(	0 == strncmp(m_szSendBuf,"QUIT", 4) ||
			0 == strncmp(m_szSendBuf,"Quit", 4) ||
			0 == strncmp(m_szSendBuf,"quit", 4) )
		{
			m_bApplicationRunning = false;		// Signal to Receive Thread to stop its operation
			cout << endl << "Quiting" << endl;
			CloseSocketsAndExit();
		}
		iNumberOfBytesSent = SendMessage();
		cout << endl << "SendThread: Sent " << m_szSendBuf << " (" << iNumberOfBytesSent << " bytes)" << endl;
	}
}

int SendMessage(void)
{
	int iLength = strlen(m_szSendBuf);

	// Send a UDP datagram containing the repy message
	int iBytesSent = sendto(m_IPC_SendSOCKET, (char FAR *)&m_szSendBuf, iLength, 0,
		(const struct sockaddr FAR *)&m_SendSockAddr, sizeof(m_SendSockAddr));
	if(INVALID_SOCKET == iBytesSent)
	{
		cout << "SendThread: sendto() Failed!" << endl;
		CloseSocketsAndExit();
	}
	return iBytesSent;
}
// *********************************** End of Send Thread code ***********************************

void CloseSocketsAndExit()
{
 	if(0 != m_IPC_ReceiveSOCKET)
	{
		closesocket(m_IPC_ReceiveSOCKET);	// If the socket was open, Close the socket
	}
 	if(0 != m_IPC_SendSOCKET)
	{
		closesocket(m_IPC_SendSOCKET);		// If the socket was open, Close the socket
	}
	WSACleanup();	
	exit(0);
}